@@ -1,19 +1,25 @@
-from sympy.core import sympify
+from itertools import product
+from typing import Tuple as tTuple
+
 from sympy.core.add import Add
 from sympy.core.cache import cacheit
-from sympy.core.function import (
-    Function, ArgumentIndexError, _coeff_isneg,
-    expand_mul, FunctionClass, PoleError)
+from sympy.core.expr import Expr
+from sympy.core.function import (Function, ArgumentIndexError, expand_log,
+    expand_mul, FunctionClass, PoleError, expand_multinomial, expand_complex)
 from sympy.core.logic import fuzzy_and, fuzzy_not, fuzzy_or
 from sympy.core.mul import Mul
-from sympy.core.numbers import Integer, Rational
+from sympy.core.numbers import Integer, Rational, pi, I
 from sympy.core.parameters import global_parameters
 from sympy.core.power import Pow
+from sympy.core.relational import Ge
 from sympy.core.singleton import S
 from sympy.core.symbol import Wild, Dummy
+from sympy.core.sympify import sympify
 from sympy.functions.combinatorial.factorials import factorial
+from sympy.functions.elementary.complexes import arg, unpolarify, im, re, Abs
 from sympy.functions.elementary.miscellaneous import sqrt
 from sympy.ntheory import multiplicity, perfect_power
+from sympy.ntheory.factor_ import factorint
 
 # NOTE IMPORTANT
 # The series expansion code in this file is an important part of the gruntz
@@ -49,7 +55,7 @@ class ExpBase(Function):
         Examples
         ========
 
-        >>> from sympy.functions import exp
+        >>> from sympy import exp
         >>> from sympy.abc import x
         >>> exp(-x).as_numer_denom()
         (1, exp(x))
@@ -58,10 +64,12 @@ class ExpBase(Function):
         """
         # this should be the same as Pow.as_numer_denom wrt
         # exponent handling
+        if not self.is_commutative:
+            return self, S.One
         exp = self.exp
         neg_exp = exp.is_negative
         if not neg_exp and not (-exp).is_negative:
-            neg_exp = _coeff_isneg(exp)
+            neg_exp = exp.could_extract_minus_sign()
         if neg_exp:
             return S.One, self.func(-exp)
         return self, S.One
@@ -119,7 +127,8 @@ class ExpBase(Function):
         return Pow._eval_power(Pow(b, e, evaluate=False), other)
 
     def _eval_expand_power_exp(self, **hints):
-        from sympy import Sum, Product
+        from sympy.concrete.products import Product
+        from sympy.concrete.summations import Sum
         arg = self.args[0]
         if arg.is_Add and arg.is_commutative:
             return Mul.fromiter(self.func(x) for x in arg.args)
@@ -130,7 +139,7 @@ class ExpBase(Function):
 
 class exp_polar(ExpBase):
     r"""
-    Represent a 'polar number' (see g-function Sphinx documentation).
+    Represent a *polar number* (see g-function Sphinx documentation).
 
     Explanation
     ===========
@@ -145,7 +154,7 @@ class exp_polar(ExpBase):
 
     >>> from sympy import exp_polar, pi, I, exp
 
-    The main difference is that polar numbers don't "wrap around" at `2 \pi`:
+    The main difference is that polar numbers do not "wrap around" at `2 \pi`:
 
     >>> exp(2*pi*I)
     1
@@ -170,12 +179,10 @@ class exp_polar(ExpBase):
     is_comparable = False  # cannot be evalf'd
 
     def _eval_Abs(self):   # Abs is never a polar number
-        from sympy.functions.elementary.complexes import re
         return exp(re(self.args[0]))
 
     def _eval_evalf(self, prec):
         """ Careful! any evalf of polar numbers is flaky """
-        from sympy import im, pi, re
         i = im(self.args[0])
         try:
             bad = (i <= -pi or i > pi)
@@ -217,9 +224,8 @@ class exp(ExpBase, metaclass=ExpMeta):
     Examples
     ========
 
-    >>> from sympy.functions import exp
+    >>> from sympy import exp, I, pi
     >>> from sympy.abc import x
-    >>> from sympy import I, pi
     >>> exp(x)
     exp(x)
     >>> exp(x).diff(x)
@@ -251,11 +257,11 @@ class exp(ExpBase, metaclass=ExpMeta):
         from sympy.assumptions import ask, Q
         arg = self.args[0]
         if arg.is_Mul:
-            Ioo = S.ImaginaryUnit*S.Infinity
+            Ioo = I*S.Infinity
             if arg in [Ioo, -Ioo]:
                 return S.NaN
 
-            coeff = arg.as_coefficient(S.Pi*S.ImaginaryUnit)
+            coeff = arg.as_coefficient(pi*I)
             if coeff:
                 if ask(Q.integer(2*coeff)):
                     if ask(Q.even(coeff)):
@@ -263,16 +269,16 @@ class exp(ExpBase, metaclass=ExpMeta):
                     elif ask(Q.odd(coeff)):
                         return S.NegativeOne
                     elif ask(Q.even(coeff + S.Half)):
-                        return -S.ImaginaryUnit
+                        return -I
                     elif ask(Q.odd(coeff + S.Half)):
-                        return S.ImaginaryUnit
+                        return I
 
     @classmethod
     def eval(cls, arg):
         from sympy.calculus import AccumBounds
+        from sympy.matrices.matrixbase import MatrixBase
         from sympy.sets.setexpr import SetExpr
-        from sympy.matrices.matrices import MatrixBase
-        from sympy import im, logcombine, re
+        from sympy.simplify.simplify import logcombine
         if isinstance(arg, MatrixBase):
             return arg.exp()
         elif global_parameters.exp_is_pow:
@@ -297,7 +303,7 @@ class exp(ExpBase, metaclass=ExpMeta):
         elif isinstance(arg, SetExpr):
             return arg._eval_func(cls)
         elif arg.is_Mul:
-            coeff = arg.as_coefficient(S.Pi*S.ImaginaryUnit)
+            coeff = arg.as_coefficient(pi*I)
             if coeff:
                 if (2*coeff).is_integer:
                     if coeff.is_even:
@@ -305,15 +311,15 @@ class exp(ExpBase, metaclass=ExpMeta):
                     elif coeff.is_odd:
                         return S.NegativeOne
                     elif (coeff + S.Half).is_even:
-                        return -S.ImaginaryUnit
+                        return -I
                     elif (coeff + S.Half).is_odd:
-                        return S.ImaginaryUnit
+                        return I
                 elif coeff.is_Rational:
                     ncoeff = coeff % 2 # restrict to [0, 2pi)
                     if ncoeff > 1: # restrict to (-pi, pi]
                         ncoeff -= 2
                     if ncoeff != coeff:
-                        return cls(ncoeff*S.Pi*S.ImaginaryUnit)
+                        return cls(ncoeff*pi*I)
 
             # Warning: code in risch.py will be very sensitive to changes
             # in this (see DifferentialExtension).
@@ -404,9 +410,8 @@ class exp(ExpBase, metaclass=ExpMeta):
         Examples
         ========
 
-        >>> from sympy import I
+        >>> from sympy import exp, I
         >>> from sympy.abc import x
-        >>> from sympy.functions import exp
         >>> exp(x).as_real_imag()
         (exp(re(x))*cos(im(x)), exp(re(x))*sin(im(x)))
         >>> exp(1).as_real_imag()
@@ -449,7 +454,7 @@ class exp(ExpBase, metaclass=ExpMeta):
         if self.args[0].is_extended_real:
             return True
         elif self.args[0].is_imaginary:
-            arg2 = -S(2) * S.ImaginaryUnit * self.args[0] / S.Pi
+            arg2 = -S(2) * I * self.args[0] / pi
             return arg2.is_even
 
     def _eval_is_complex(self):
@@ -459,25 +464,29 @@ class exp(ExpBase, metaclass=ExpMeta):
         return fuzzy_or(complex_extended_negative(self.args[0]))
 
     def _eval_is_algebraic(self):
-        if (self.exp / S.Pi / S.ImaginaryUnit).is_rational:
+        if (self.exp / pi / I).is_rational:
             return True
         if fuzzy_not(self.exp.is_zero):
             if self.exp.is_algebraic:
                 return False
-            elif (self.exp / S.Pi).is_rational:
+            elif (self.exp / pi).is_rational:
                 return False
 
     def _eval_is_extended_positive(self):
         if self.exp.is_extended_real:
-            return not self.args[0] is S.NegativeInfinity
+            return self.args[0] is not S.NegativeInfinity
         elif self.exp.is_imaginary:
-            arg2 = -S.ImaginaryUnit * self.args[0] / S.Pi
+            arg2 = -I * self.args[0] / pi
             return arg2.is_even
 
     def _eval_nseries(self, x, n, logx, cdir=0):
         # NOTE Please see the comment at the beginning of this file, labelled
         #      IMPORTANT.
-        from sympy import ceiling, limit, Order, powsimp, Wild, expand_complex
+        from sympy.functions.elementary.complexes import sign
+        from sympy.functions.elementary.integers import ceiling
+        from sympy.series.limits import limit
+        from sympy.series.order import Order
+        from sympy.simplify.powsimp import powsimp
         arg = self.exp
         arg_series = arg._eval_nseries(x, n=n, logx=logx)
         if arg_series.is_Order:
@@ -487,6 +496,11 @@ class exp(ExpBase, metaclass=ExpMeta):
             return Order(x**n, x)
         if arg0 is S.Infinity:
             return self
+        if arg0.is_infinite:
+            raise PoleError("Cannot expand %s around 0" % (self))
+        # checking for indecisiveness/ sign terms in arg0
+        if any(isinstance(arg, sign) for arg in arg0.args):
+            return self
         t = Dummy("t")
         nterms = n
         try:
@@ -497,6 +511,9 @@ class exp(ExpBase, metaclass=ExpMeta):
             nterms = ceiling(n/cf)
         exp_series = exp(t)._taylor(t, nterms)
         r = exp(arg0)*exp_series.subs(t, arg_series - arg0)
+        rep = {logx: log(x)} if logx is not None else {}
+        if r.subs(rep) == self:
+            return r
         if cf and cf > 1:
             r += Order((arg_series - arg0)**n, x)/x**((cf-1)*n)
         else:
@@ -506,7 +523,7 @@ class exp(ExpBase, metaclass=ExpMeta):
         # powsimp may introduce unexpanded (-1)**Rational; see PR #17201
         simplerat = lambda x: x.is_Rational and x.q in [3, 4, 6]
         w = Wild('w', properties=[simplerat])
-        r = r.replace((-1)**w, expand_complex((-1)**w))
+        r = r.replace(S.NegativeOne**w, expand_complex(S.NegativeOne**w))
         return r
 
     def _taylor(self, x, n):
@@ -519,8 +536,20 @@ class exp(ExpBase, metaclass=ExpMeta):
         return Add(*l)
 
     def _eval_as_leading_term(self, x, logx=None, cdir=0):
+        from sympy.calculus.util import AccumBounds
         arg = self.args[0].cancel().as_leading_term(x, logx=logx)
         arg0 = arg.subs(x, 0)
+        if arg is S.NaN:
+            return S.NaN
+        if isinstance(arg0, AccumBounds):
+            # This check addresses a corner case involving AccumBounds.
+            # if isinstance(arg, AccumBounds) is True, then arg0 can either be 0,
+            # AccumBounds(-oo, 0) or AccumBounds(-oo, oo).
+            # Check out function: test_issue_18473() in test_exponential.py and
+            # test_limits.py for more information.
+            if re(cdir) < S.Zero:
+                return exp(-arg0)
+            return exp(arg0)
         if arg0 is S.NaN:
             arg0 = arg.limit(x, 0)
         if arg0.is_infinite is False:
@@ -528,27 +557,25 @@ class exp(ExpBase, metaclass=ExpMeta):
         raise PoleError("Cannot expand %s around 0" % (self))
 
     def _eval_rewrite_as_sin(self, arg, **kwargs):
-        from sympy import sin
-        I = S.ImaginaryUnit
-        return sin(I*arg + S.Pi/2) - I*sin(I*arg)
+        from sympy.functions.elementary.trigonometric import sin
+        return sin(I*arg + pi/2) - I*sin(I*arg)
 
     def _eval_rewrite_as_cos(self, arg, **kwargs):
-        from sympy import cos
-        I = S.ImaginaryUnit
-        return cos(I*arg) + I*cos(I*arg + S.Pi/2)
+        from sympy.functions.elementary.trigonometric import cos
+        return cos(I*arg) + I*cos(I*arg + pi/2)
 
     def _eval_rewrite_as_tanh(self, arg, **kwargs):
-        from sympy import tanh
+        from sympy.functions.elementary.hyperbolic import tanh
         return (1 + tanh(arg/2))/(1 - tanh(arg/2))
 
     def _eval_rewrite_as_sqrt(self, arg, **kwargs):
         from sympy.functions.elementary.trigonometric import sin, cos
         if arg.is_Mul:
-            coeff = arg.coeff(S.Pi*S.ImaginaryUnit)
+            coeff = arg.coeff(pi*I)
             if coeff and coeff.is_number:
-                cosine, sine = cos(S.Pi*coeff), sin(S.Pi*coeff)
+                cosine, sine = cos(pi*coeff), sin(pi*coeff)
                 if not isinstance(cosine, cos) and not isinstance (sine, sin):
-                    return cosine + S.ImaginaryUnit*sine
+                    return cosine + I*sine
 
     def _eval_rewrite_as_Pow(self, arg, **kwargs):
         if arg.is_Mul:
@@ -558,20 +585,20 @@ class exp(ExpBase, metaclass=ExpMeta):
 
 
 def match_real_imag(expr):
-    """
-    Try to match expr with a + b*I for real a and b.
+    r"""
+    Try to match expr with $a + Ib$ for real $a$ and $b$.
 
     ``match_real_imag`` returns a tuple containing the real and imaginary
-    parts of expr or (None, None) if direct matching is not possible. Contrary
-    to ``re()``, ``im()``, ``as_real_imag()``, this helper won't force things
+    parts of expr or ``(None, None)`` if direct matching is not possible. Contrary
+    to :func:`~.re()`, :func:`~.im()``, and ``as_real_imag()``, this helper will not force things
     by returning expressions themselves containing ``re()`` or ``im()`` and it
-    doesn't expand its argument either.
+    does not expand its argument either.
 
     """
-    r_, i_ = expr.as_independent(S.ImaginaryUnit, as_Add=True)
+    r_, i_ = expr.as_independent(I, as_Add=True)
     if i_ == 0 and r_.is_real:
         return (r_, i_)
-    i_ = i_.as_coefficient(S.ImaginaryUnit)
+    i_ = i_.as_coefficient(I)
     if i_ and i_.is_real and r_.is_real:
         return (r_, i_)
     else:
@@ -611,6 +638,9 @@ class log(Function):
     exp
 
     """
+
+    args: tTuple[Expr]
+
     _singularities = (S.Zero, S.ComplexInfinity)
 
     def fdiff(self, argindex=1):
@@ -630,10 +660,8 @@ class log(Function):
 
     @classmethod
     def eval(cls, arg, base=None):
-        from sympy import unpolarify
         from sympy.calculus import AccumBounds
         from sympy.sets.setexpr import SetExpr
-        from sympy.functions.elementary.complexes import Abs
 
         arg = sympify(arg)
 
@@ -675,29 +703,30 @@ class log(Function):
 
         if arg.is_Pow and arg.base is S.Exp1 and arg.exp.is_extended_real:
             return arg.exp
-        I = S.ImaginaryUnit
         if isinstance(arg, exp) and arg.exp.is_extended_real:
             return arg.exp
         elif isinstance(arg, exp) and arg.exp.is_number:
             r_, i_ = match_real_imag(arg.exp)
             if i_ and i_.is_comparable:
-                i_ %= 2*S.Pi
-                if i_ > S.Pi:
-                    i_ -= 2*S.Pi
+                i_ %= 2*pi
+                if i_ > pi:
+                    i_ -= 2*pi
                 return r_ + expand_mul(i_ * I, deep=False)
         elif isinstance(arg, exp_polar):
             return unpolarify(arg.exp)
         elif isinstance(arg, AccumBounds):
             if arg.min.is_positive:
                 return AccumBounds(log(arg.min), log(arg.max))
+            elif arg.min.is_zero:
+                return AccumBounds(S.NegativeInfinity, log(arg.max))
             else:
-                return
+                return S.NaN
         elif isinstance(arg, SetExpr):
             return arg._eval_func(cls)
 
         if arg.is_number:
             if arg.is_negative:
-                return S.Pi * I + cls(-arg)
+                return pi * I + cls(-arg)
             elif arg is S.ComplexInfinity:
                 return S.ComplexInfinity
             elif arg is S.Exp1:
@@ -717,9 +746,9 @@ class log(Function):
                     return S.Infinity
                 elif coeff.is_Rational:
                     if coeff.is_nonnegative:
-                        return S.Pi * I * S.Half + cls(coeff)
+                        return pi * I * S.Half + cls(coeff)
                     else:
-                        return -S.Pi * I * S.Half + cls(-coeff)
+                        return -pi * I * S.Half + cls(-coeff)
 
         if arg.is_number and arg.is_algebraic:
             # Match arg = coeff*(r_ + i_*I) with coeff>0, r_ and i_ real.
@@ -733,48 +762,27 @@ class log(Function):
             if coeff.is_real and i_ and i_.is_real and r_.is_real:
                 if r_.is_zero:
                     if i_.is_positive:
-                        return S.Pi * I * S.Half + cls(coeff * i_)
+                        return pi * I * S.Half + cls(coeff * i_)
                     elif i_.is_negative:
-                        return -S.Pi * I * S.Half + cls(coeff * -i_)
+                        return -pi * I * S.Half + cls(coeff * -i_)
                 else:
                     from sympy.simplify import ratsimp
                     # Check for arguments involving rational multiples of pi
                     t = (i_/r_).cancel()
                     t1 = (-t).cancel()
-                    atan_table = {
-                        # first quadrant only
-                        sqrt(3): S.Pi/3,
-                        1: S.Pi/4,
-                        sqrt(5 - 2*sqrt(5)): S.Pi/5,
-                        sqrt(2)*sqrt(5 - sqrt(5))/(1 + sqrt(5)): S.Pi/5,
-                        sqrt(5 + 2*sqrt(5)): S.Pi*Rational(2, 5),
-                        sqrt(2)*sqrt(sqrt(5) + 5)/(-1 + sqrt(5)): S.Pi*Rational(2, 5),
-                        sqrt(3)/3: S.Pi/6,
-                        sqrt(2) - 1: S.Pi/8,
-                        sqrt(2 - sqrt(2))/sqrt(sqrt(2) + 2): S.Pi/8,
-                        sqrt(2) + 1: S.Pi*Rational(3, 8),
-                        sqrt(sqrt(2) + 2)/sqrt(2 - sqrt(2)): S.Pi*Rational(3, 8),
-                        sqrt(1 - 2*sqrt(5)/5): S.Pi/10,
-                        (-sqrt(2) + sqrt(10))/(2*sqrt(sqrt(5) + 5)): S.Pi/10,
-                        sqrt(1 + 2*sqrt(5)/5): S.Pi*Rational(3, 10),
-                        (sqrt(2) + sqrt(10))/(2*sqrt(5 - sqrt(5))): S.Pi*Rational(3, 10),
-                        2 - sqrt(3): S.Pi/12,
-                        (-1 + sqrt(3))/(1 + sqrt(3)): S.Pi/12,
-                        2 + sqrt(3): S.Pi*Rational(5, 12),
-                        (1 + sqrt(3))/(-1 + sqrt(3)): S.Pi*Rational(5, 12)
-                    }
+                    atan_table = _log_atan_table()
                     if t in atan_table:
                         modulus = ratsimp(coeff * Abs(arg_))
                         if r_.is_positive:
                             return cls(modulus) + I * atan_table[t]
                         else:
-                            return cls(modulus) + I * (atan_table[t] - S.Pi)
+                            return cls(modulus) + I * (atan_table[t] - pi)
                     elif t1 in atan_table:
                         modulus = ratsimp(coeff * Abs(arg_))
                         if r_.is_positive:
                             return cls(modulus) + I * (-atan_table[t1])
                         else:
-                            return cls(modulus) + I * (S.Pi - atan_table[t1])
+                            return cls(modulus) + I * (pi - atan_table[t1])
 
     def as_base_exp(self):
         """
@@ -788,7 +796,7 @@ class log(Function):
         r"""
         Returns the next term in the Taylor series expansion of `\log(1+x)`.
         """
-        from sympy import powsimp
+        from sympy.simplify.powsimp import powsimp
         if n < 0:
             return S.Zero
         x = sympify(x)
@@ -801,7 +809,6 @@ class log(Function):
         return (1 - 2*(n % 2)) * x**(n + 1)/(n + 1)
 
     def _eval_expand_log(self, deep=True, **hints):
-        from sympy import unpolarify, expand_log, factorint
         from sympy.concrete import Sum, Product
         force = hints.get('force', False)
         factor = hints.get('factor', False)
@@ -876,9 +883,8 @@ class log(Function):
         Examples
         ========
 
-        >>> from sympy import I
+        >>> from sympy import I, log
         >>> from sympy.abc import x
-        >>> from sympy.functions import log
         >>> log(x).as_real_imag()
         (log(Abs(x)), arg(x))
         >>> log(I).as_real_imag()
@@ -889,19 +895,18 @@ class log(Function):
         (log(Abs(x)), arg(I*x))
 
         """
-        from sympy import Abs, arg
         sarg = self.args[0]
         if deep:
             sarg = self.args[0].expand(deep, **hints)
-        abs = Abs(sarg)
-        if abs == sarg:
+        sarg_abs = Abs(sarg)
+        if sarg_abs == sarg:
             return self, S.Zero
-        arg = arg(sarg)
+        sarg_arg = arg(sarg)
         if hints.get('log', False):  # Expand the log
             hints['complex'] = False
-            return (log(abs).expand(deep, **hints), arg)
+            return (log(sarg_abs).expand(deep, **hints), sarg_arg)
         else:
-            return log(abs), arg
+            return log(sarg_abs), sarg_arg
 
     def _eval_is_rational(self):
         s = self.func(*self.args)
@@ -949,19 +954,25 @@ class log(Function):
     def _eval_nseries(self, x, n, logx, cdir=0):
         # NOTE Please see the comment at the beginning of this file, labelled
         #      IMPORTANT.
-        from sympy import im, cancel, I, Order, logcombine
-        from itertools import product
-        if not logx:
-            logx = log(x)
+        from sympy.series.order import Order
+        from sympy.simplify.simplify import logcombine
+        from sympy.core.symbol import Dummy
+
         if self.args[0] == x:
-            return logx
+            return log(x) if logx is None else logx
         arg = self.args[0]
+        t = Dummy('t', positive=True)
+        if cdir == 0:
+            cdir = 1
+        z = arg.subs(x, cdir*t)
+
         k, l = Wild("k"), Wild("l")
-        r = arg.match(k*x**l)
+        r = z.match(k*t**l)
         if r is not None:
             k, l = r[k], r[l]
-            if l != 0 and not l.has(x) and not k.has(x):
-                r = log(k) + l*logx  # XXX true regardless of assumptions?
+            if l != 0 and not l.has(t) and not k.has(t):
+                r = l*log(x) if logx is None else l*logx
+                r += log(k) - l*log(cdir) # XXX true regardless of assumptions?
                 return r
 
         def coeff_exp(term, x):
@@ -980,22 +991,40 @@ class log(Function):
 
         # TODO new and probably slow
         try:
-            a, b = arg.leadterm(x)
-            s = arg.nseries(x, n=n+b, logx=logx)
+            a, b = z.leadterm(t, logx=logx, cdir=1)
         except (ValueError, NotImplementedError, PoleError):
-            s = arg.nseries(x, n=n, logx=logx)
+            s = z._eval_nseries(t, n=n, logx=logx, cdir=1)
             while s.is_Order:
                 n += 1
-                s = arg.nseries(x, n=n, logx=logx)
-        a, b = s.removeO().leadterm(x)
-        p = cancel(s/(a*x**b) - 1).expand().powsimp()
+                s = z._eval_nseries(t, n=n, logx=logx, cdir=1)
+            try:
+                a, b = s.removeO().leadterm(t, cdir=1)
+            except ValueError:
+                a, b = s.removeO().as_leading_term(t, cdir=1), S.Zero
+
+        p = (z/(a*t**b) - 1)._eval_nseries(t, n=n, logx=logx, cdir=1)
         if p.has(exp):
             p = logcombine(p)
         if isinstance(p, Order):
             n = p.getn()
-        _, d = coeff_exp(p, x)
+        _, d = coeff_exp(p, t)
+        logx = log(x) if logx is None else logx
+
         if not d.is_positive:
-            return log(a) + b*logx + Order(x**n, x)
+            res = log(a) - b*log(cdir) + b*logx
+            _res = res
+            logflags = {"deep": True, "log": True, "mul": False, "power_exp": False,
+                "power_base": False, "multinomial": False, "basic": False, "force": True,
+                "factor": False}
+            expr = self.expand(**logflags)
+            if (not a.could_extract_minus_sign() and
+                logx.could_extract_minus_sign()):
+                _res = _res.subs(-logx, -log(x)).expand(**logflags)
+            else:
+                _res = _res.subs(logx, log(x)).expand(**logflags)
+            if _res == expr:
+                return res
+            return res + Order(x**n, x)
 
         def mul(d1, d2):
             res = {}
@@ -1007,68 +1036,101 @@ class log(Function):
 
         pterms = {}
 
-        for term in Add.make_args(p):
-            co1, e1 = coeff_exp(term, x)
-            pterms[e1] = pterms.get(e1, S.Zero) + co1.removeO()
+        for term in Add.make_args(p.removeO()):
+            co1, e1 = coeff_exp(term, t)
+            pterms[e1] = pterms.get(e1, S.Zero) + co1
 
         k = S.One
         terms = {}
         pk = pterms
 
         while k*d < n:
-            coeff = -(-1)**k/k
+            coeff = -S.NegativeOne**k/k
             for ex in pk:
-                terms[ex] = terms.get(ex, S.Zero) + coeff*pk[ex]
+                _ = terms.get(ex, S.Zero) + coeff*pk[ex]
+                terms[ex] = _.nsimplify()
             pk = mul(pk, pterms)
             k += S.One
 
-        res = log(a) + b*logx
+        res = log(a) - b*log(cdir) + b*logx
         for ex in terms:
-            res += terms[ex]*x**(ex)
+            res += terms[ex]*t**(ex)
 
-        if cdir != 0:
-            cdir = self.args[0].dir(x, cdir)
-        if a.is_real and a.is_negative and im(cdir) < 0:
-            res -= 2*I*S.Pi
+        if a.is_negative and im(z) != 0:
+            from sympy.functions.special.delta_functions import Heaviside
+            for i, term in enumerate(z.lseries(t)):
+                if not term.is_real or i == 5:
+                    break
+            if i < 5:
+                coeff, _ = term.as_coeff_exponent(t)
+                res += -2*I*pi*Heaviside(-im(coeff), 0)
+
+        res = res.subs(t, x/cdir)
         return res + Order(x**n, x)
 
     def _eval_as_leading_term(self, x, logx=None, cdir=0):
-        from sympy import I, im, re
+        # NOTE
+        # Refer https://github.com/sympy/sympy/pull/23592 for more information
+        # on each of the following steps involved in this method.
         arg0 = self.args[0].together()
 
-        arg = arg0.as_leading_term(x, cdir=cdir)
-        x0 = arg0.subs(x, 0)
-        if (x0 is S.NaN and logx is None):
-            x0 = arg.limit(x, 0, dir='-' if re(cdir).is_negative else '+')
-        if x0 in (S.NegativeInfinity, S.Infinity):
-            raise PoleError("Cannot expand %s around 0" % (self))
-        if x0 == 1:
-            return (arg0 - S.One).as_leading_term(x)
-        if cdir != 0:
-            cdir = arg0.dir(x, cdir)
-        if x0.is_real and x0.is_negative and im(cdir).is_negative:
-            return self.func(x0) - 2*I*S.Pi
-        return self.func(arg)
+        # STEP 1
+        t = Dummy('t', positive=True)
+        if cdir == 0:
+            cdir = 1
+        z = arg0.subs(x, cdir*t)
+
+        # STEP 2
+        try:
+            c, e = z.leadterm(t, logx=logx, cdir=1)
+        except ValueError:
+            arg = arg0.as_leading_term(x, logx=logx, cdir=cdir)
+            return log(arg)
+        if c.has(t):
+            c = c.subs(t, x/cdir)
+            if e != 0:
+                raise PoleError("Cannot expand %s around 0" % (self))
+            return log(c)
+
+        # STEP 3
+        if c == S.One and e == S.Zero:
+            return (arg0 - S.One).as_leading_term(x, logx=logx)
+
+        # STEP 4
+        res = log(c) - e*log(cdir)
+        logx = log(x) if logx is None else logx
+        res += e*logx
+
+        # STEP 5
+        if c.is_negative and im(z) != 0:
+            from sympy.functions.special.delta_functions import Heaviside
+            for i, term in enumerate(z.lseries(t)):
+                if not term.is_real or i == 5:
+                    break
+            if i < 5:
+                coeff, _ = term.as_coeff_exponent(t)
+                res += -2*I*pi*Heaviside(-im(coeff), 0)
+        return res
 
 
 class LambertW(Function):
     r"""
-    The Lambert W function `W(z)` is defined as the inverse
-    function of `w \exp(w)` [1]_.
+    The Lambert W function $W(z)$ is defined as the inverse
+    function of $w \exp(w)$ [1]_.
 
     Explanation
     ===========
 
-    In other words, the value of `W(z)` is such that `z = W(z) \exp(W(z))`
-    for any complex number `z`.  The Lambert W function is a multivalued
-    function with infinitely many branches `W_k(z)`, indexed by
-    `k \in \mathbb{Z}`.  Each branch gives a different solution `w`
-    of the equation `z = w \exp(w)`.
+    In other words, the value of $W(z)$ is such that $z = W(z) \exp(W(z))$
+    for any complex number $z$.  The Lambert W function is a multivalued
+    function with infinitely many branches $W_k(z)$, indexed by
+    $k \in \mathbb{Z}$.  Each branch gives a different solution $w$
+    of the equation $z = w \exp(w)$.
 
     The Lambert W function has two partially real branches: the
-    principal branch (`k = 0`) is real for real `z > -1/e`, and the
-    `k = -1` branch is real for `-1/e < z < 0`. All branches except
-    `k = 0` have a logarithmic singularity at `z = 0`.
+    principal branch ($k = 0$) is real for real $z > -1/e$, and the
+    $k = -1$ branch is real for $-1/e < z < 0$. All branches except
+    $k = 0$ have a logarithmic singularity at $z = 0$.
 
     Examples
     ========
@@ -1100,27 +1162,37 @@ class LambertW(Function):
                 return S.Zero
             if x is S.Exp1:
                 return S.One
-            if x == -1/S.Exp1:
-                return S.NegativeOne
+            w = Wild('w')
+            # W(x*log(x)) = log(x) for x >= 1/e
+            # e.g., W(-1/e) = -1, W(2*log(2)) = log(2)
+            result = x.match(w*log(w))
+            if result is not None and Ge(result[w]*S.Exp1, S.One) is S.true:
+                return log(result[w])
             if x == -log(2)/2:
                 return -log(2)
-            if x == 2*log(2):
-                return log(2)
-            if x == -S.Pi/2:
-                return S.ImaginaryUnit*S.Pi/2
+            # W(x**(x+1)*log(x)) = x*log(x) for x > 0
+            # e.g., W(81*log(3)) = 3*log(3)
+            result = x.match(w**(w+1)*log(w))
+            if result is not None and result[w].is_positive is True:
+                return result[w]*log(result[w])
+            # W(e**(1/n)/n) = 1/n
+            # e.g., W(sqrt(e)/2) = 1/2
+            result = x.match(S.Exp1**(1/w)/w)
+            if result is not None:
+                return 1 / result[w]
+            if x == -pi/2:
+                return I*pi/2
             if x == exp(1 + S.Exp1):
                 return S.Exp1
             if x is S.Infinity:
                 return S.Infinity
-            if x.is_zero:
-                return S.Zero
 
         if fuzzy_not(k.is_zero):
             if x.is_zero:
                 return S.NegativeInfinity
         if k is S.NegativeOne:
-            if x == -S.Pi/2:
-                return -S.ImaginaryUnit*S.Pi/2
+            if x == -pi/2:
+                return -I*pi/2
             elif x == -1/S.Exp1:
                 return S.NegativeOne
             elif x == -2*exp(-2):
@@ -1183,9 +1255,10 @@ class LambertW(Function):
 
     def _eval_nseries(self, x, n, logx, cdir=0):
         if len(self.args) == 1:
-            from sympy import Order, ceiling, expand_multinomial
+            from sympy.functions.elementary.integers import ceiling
+            from sympy.series.order import Order
             arg = self.args[0].nseries(x, n=n, logx=logx)
-            lt = arg.compute_leading_term(x, logx=logx)
+            lt = arg.as_leading_term(x, logx=logx)
             lte = 1
             if lt.is_Pow:
                 lte = lt.exp
@@ -1202,8 +1275,32 @@ class LambertW(Function):
     def _eval_is_zero(self):
         x = self.args[0]
         if len(self.args) == 1:
-            k = S.Zero
+            return x.is_zero
         else:
-            k = self.args[1]
-        if x.is_zero and k.is_zero:
-            return True
+            return fuzzy_and([x.is_zero, self.args[1].is_zero])
+
+
+@cacheit
+def _log_atan_table():
+    return {
+        # first quadrant only
+        sqrt(3): pi / 3,
+        1: pi / 4,
+        sqrt(5 - 2 * sqrt(5)): pi / 5,
+        sqrt(2) * sqrt(5 - sqrt(5)) / (1 + sqrt(5)): pi / 5,
+        sqrt(5 + 2 * sqrt(5)): pi * Rational(2, 5),
+        sqrt(2) * sqrt(sqrt(5) + 5) / (-1 + sqrt(5)): pi * Rational(2, 5),
+        sqrt(3) / 3: pi / 6,
+        sqrt(2) - 1: pi / 8,
+        sqrt(2 - sqrt(2)) / sqrt(sqrt(2) + 2): pi / 8,
+        sqrt(2) + 1: pi * Rational(3, 8),
+        sqrt(sqrt(2) + 2) / sqrt(2 - sqrt(2)): pi * Rational(3, 8),
+        sqrt(1 - 2 * sqrt(5) / 5): pi / 10,
+        (-sqrt(2) + sqrt(10)) / (2 * sqrt(sqrt(5) + 5)): pi / 10,
+        sqrt(1 + 2 * sqrt(5) / 5): pi * Rational(3, 10),
+        (sqrt(2) + sqrt(10)) / (2 * sqrt(5 - sqrt(5))): pi * Rational(3, 10),
+        2 - sqrt(3): pi / 12,
+        (-1 + sqrt(3)) / (1 + sqrt(3)): pi / 12,
+        2 + sqrt(3): pi * Rational(5, 12),
+        (1 + sqrt(3)) / (-1 + sqrt(3)): pi * Rational(5, 12)
+    }
