@@ -1,20 +1,19 @@
 from types import FunctionType
 
+from sympy.core.cache import cacheit
 from sympy.core.numbers import Float, Integer
 from sympy.core.singleton import S
 from sympy.core.symbol import uniquely_named_symbol
 from sympy.core.mul import Mul
 from sympy.polys import PurePoly, cancel
-from sympy.simplify.simplify import (simplify as _simplify,
-    dotprodsimp as _dotprodsimp)
-from sympy import sympify
 from sympy.functions.combinatorial.numbers import nC
 from sympy.polys.matrices.domainmatrix import DomainMatrix
+from sympy.polys.matrices.ddm import DDM
 
-from .common import NonSquareMatrixError
+from .exceptions import NonSquareMatrixError
 from .utilities import (
     _get_intermediate_simp, _get_intermediate_simp_bool,
-    _iszero, _is_zero_after_expand_mul)
+    _iszero, _is_zero_after_expand_mul, _dotprodsimp, _simplify)
 
 
 def _find_reasonable_pivot(col, iszerofunc=_iszero, simpfunc=_simplify):
@@ -22,7 +21,7 @@ def _find_reasonable_pivot(col, iszerofunc=_iszero, simpfunc=_simplify):
     suitable for a pivot.  If ``col`` consists only of
     Floats, the pivot with the largest norm is returned.
     Otherwise, the first element where ``iszerofunc`` returns
-    False is used.  If ``iszerofunc`` doesn't return false,
+    False is used.  If ``iszerofunc`` does not return false,
     items are simplified and retested until a suitable
     pivot is found.
 
@@ -82,7 +81,7 @@ def _find_reasonable_pivot(col, iszerofunc=_iszero, simpfunc=_simplify):
             continue
         simped = simpfunc(x)
         is_zero = iszerofunc(simped)
-        if is_zero == True or is_zero == False:
+        if is_zero in (True, False):
             newly_determined.append((i, simped))
         if is_zero == False:
             return (i, simped, False, newly_determined)
@@ -306,8 +305,8 @@ def _adjugate(M, method="berkowitz"):
     ==========
 
     method : string, optional
-        Method to use to find the cofactors, can be "bareiss", "berkowitz" or
-        "lu".
+        Method to use to find the cofactors, can be "bareiss", "berkowitz",
+        "bird", "laplace" or "lu".
 
     Examples
     ========
@@ -323,7 +322,7 @@ def _adjugate(M, method="berkowitz"):
     ========
 
     cofactor_matrix
-    sympy.matrices.common.MatrixCommon.transpose
+    sympy.matrices.matrixbase.MatrixBase.transpose
     """
 
     return M.cofactor_matrix(method=method).transpose()
@@ -403,18 +402,40 @@ def _charpoly(M, x='lambda', simplify=_simplify):
 
     if not M.is_square:
         raise NonSquareMatrixError()
-    if M.is_lower or M.is_upper:
-        diagonal_elements = M.diagonal()
-        x = uniquely_named_symbol(x, diagonal_elements, modify=lambda s: '_' + s)
-        m = 1
-        for i in diagonal_elements:
-            m = m * (x - simplify(i))
-        return PurePoly(m, x)
 
-    berk_vector = _berkowitz_vector(M)
-    x = uniquely_named_symbol(x, berk_vector, modify=lambda s: '_' + s)
+    # Use DomainMatrix. We are already going to convert this to a Poly so there
+    # is no need to worry about expanding powers etc. Also since this algorithm
+    # does not require division or zero detection it is fine to use EX.
+    #
+    # M.to_DM() will fall back on EXRAW rather than EX. EXRAW is a lot faster
+    # for elementary arithmetic because it does not call cancel for each
+    # operation but it generates large unsimplified results that are slow in
+    # the subsequent call to simplify. Using EX instead is faster overall
+    # but at least in some cases EXRAW+simplify gives a simpler result so we
+    # preserve that existing behaviour of charpoly for now...
+    dM = M.to_DM()
 
-    return PurePoly([simplify(a) for a in berk_vector], x)
+    K = dM.domain
+
+    cp = dM.charpoly()
+
+    x = uniquely_named_symbol(x, [M], modify=lambda s: '_' + s)
+
+    if K.is_EXRAW or simplify is not _simplify:
+        # XXX: Converting back to Expr is expensive. We only do it if the
+        # caller supplied a custom simplify function for backwards
+        # compatibility or otherwise if the domain was EX. For any other domain
+        # there should be no benefit in simplifying at this stage because Poly
+        # will put everything into canonical form anyway.
+        berk_vector = [K.to_sympy(c) for c in cp]
+        berk_vector = [simplify(a) for a in berk_vector]
+        p = PurePoly(berk_vector, x)
+
+    else:
+        # Convert from the list of domain elements directly to Poly.
+        p = PurePoly(cp, x, domain=K)
+
+    return p
 
 
 def _cofactor(M, i, j, method="berkowitz"):
@@ -424,8 +445,8 @@ def _cofactor(M, i, j, method="berkowitz"):
     ==========
 
     method : string, optional
-        Method to use to find the cofactors, can be "bareiss", "berkowitz" or
-        "lu".
+        Method to use to find the cofactors, can be "bareiss", "berkowitz",
+        "bird", "laplace" or "lu".
 
     Examples
     ========
@@ -446,7 +467,7 @@ def _cofactor(M, i, j, method="berkowitz"):
     if not M.is_square or M.rows < 1:
         raise NonSquareMatrixError()
 
-    return (-1)**((i + j) % 2) * M.minor(i, j, method)
+    return S.NegativeOne**((i + j) % 2) * M.minor(i, j, method)
 
 
 def _cofactor_matrix(M, method="berkowitz"):
@@ -456,8 +477,8 @@ def _cofactor_matrix(M, method="berkowitz"):
     ==========
 
     method : string, optional
-        Method to use to find the cofactors, can be "bareiss", "berkowitz" or
-        "lu".
+        Method to use to find the cofactors, can be "bareiss", "berkowitz",
+        "bird", "laplace" or "lu".
 
     Examples
     ========
@@ -477,7 +498,7 @@ def _cofactor_matrix(M, method="berkowitz"):
     minor_submatrix
     """
 
-    if not M.is_square or M.rows < 1:
+    if not M.is_square:
         raise NonSquareMatrixError()
 
     return M._new(M.rows, M.cols,
@@ -513,7 +534,7 @@ def _per(M):
     ==========
 
     .. [1] Prof. Frank Ben's notes: https://math.berkeley.edu/~bernd/ban275.pdf
-    .. [2] Wikipedia article on Permanent: https://en.wikipedia.org/wiki/Permanent_(mathematics)
+    .. [2] Wikipedia article on Permanent: https://en.wikipedia.org/wiki/Permanent_%28mathematics%29
     .. [3] https://reference.wolfram.com/language/ref/Permanent.html
     .. [4] Permanent of a rectangular matrix : https://arxiv.org/pdf/0904.3251.pdf
     """
@@ -534,10 +555,9 @@ def _per(M):
         prod = 1
         sub_len = len(subset)
         for i in range(m):
-             prod *= sum([M[i, j] for j in subset])
-        perm += prod * (-1)**sub_len * nC(n - sub_len, m - sub_len)
-    perm *= (-1)**m
-    perm = sympify(perm)
+             prod *= sum(M[i, j] for j in subset)
+        perm += prod * S.NegativeOne**sub_len * nC(n - sub_len, m - sub_len)
+    perm *= S.NegativeOne**m
     return perm.simplify()
 
 def _det_DOM(M):
@@ -573,6 +593,10 @@ def _det(M, method="bareiss", iszerofunc=None):
 
         If it is set to ``'berkowitz'``, Berkowitz' algorithm will be used.
 
+        If it is set to ``'bird'``, Bird's algorithm will be used [1]_.
+
+        If it is set to ``'laplace'``, Laplace's algorithm will be used.
+
         Otherwise, if it is set to ``'lu'``, LU decomposition will be used.
 
         .. note::
@@ -621,6 +645,13 @@ def _det(M, method="bareiss", iszerofunc=None):
     True
     >>> M.det(method="domain-ge")
     -2
+
+    References
+    ==========
+
+    .. [1] Bird, R. S. (2011). A simple division-free algorithm for computing
+           determinants. Inf. Process. Lett., 111(21), 1072-1074. doi:
+           10.1016/j.ipl.2011.08.006
     """
 
     # sanitize `method`
@@ -631,7 +662,8 @@ def _det(M, method="bareiss", iszerofunc=None):
     elif method == "det_lu":
         method = "lu"
 
-    if method not in ("bareiss", "berkowitz", "lu", "domain-ge"):
+    if method not in ("bareiss", "berkowitz", "lu", "domain-ge", "bird",
+                      "laplace"):
         raise ValueError("Determinant method '%s' unrecognized" % method)
 
     if iszerofunc is None:
@@ -664,7 +696,7 @@ def _det(M, method="bareiss", iszerofunc=None):
 
     dets = []
     for b in M.strongly_connected_components():
-        if method == "domain-ge": # uses DomainMatrix to evalute determinant
+        if method == "domain-ge": # uses DomainMatrix to evaluate determinant
             det = _det_DOM(M[b, b])
         elif method == "bareiss":
             det = M[b, b]._eval_det_bareiss(iszerofunc=iszerofunc)
@@ -672,6 +704,10 @@ def _det(M, method="bareiss", iszerofunc=None):
             det = M[b, b]._eval_det_berkowitz()
         elif method == "lu":
             det = M[b, b]._eval_det_lu(iszerofunc=iszerofunc)
+        elif method == "bird":
+            det = M[b, b]._eval_det_bird()
+        elif method == "laplace":
+            det = M[b, b]._eval_det_laplace()
         dets.append(det)
     return Mul(*dets)
 
@@ -717,7 +753,7 @@ def _det_bareiss(M, iszerofunc=_is_zero_after_expand_mul):
         sign = (-1) ** (pivot_pos % 2)
 
         # we want every row but the pivot row and every column
-        rows = list(i for i in range(mat.rows) if i != pivot_pos)
+        rows = [i for i in range(mat.rows) if i != pivot_pos]
         cols = list(range(mat.cols))
         tmp_mat = mat.extract(rows, cols)
 
@@ -823,6 +859,89 @@ def _det_LU(M, iszerofunc=_iszero, simpfunc=None):
     return det
 
 
+@cacheit
+def __det_laplace(M):
+    """Compute the determinant of a matrix using Laplace expansion.
+
+    This is a recursive function, and it should not be called directly.
+    Use _det_laplace() instead. The reason for splitting this function
+    into two is to allow caching of determinants of submatrices. While
+    one could also define this function inside _det_laplace(), that
+    would remove the advantage of using caching in Cramer Solve.
+    """
+    n = M.shape[0]
+    if n == 1:
+        return M[0]
+    elif n == 2:
+        return M[0, 0] * M[1, 1] - M[0, 1] * M[1, 0]
+    else:
+        return sum((-1) ** i * M[0, i] *
+                   __det_laplace(M.minor_submatrix(0, i)) for i in range(n))
+
+
+def _det_laplace(M):
+    """Compute the determinant of a matrix using Laplace expansion.
+
+    While Laplace expansion is not the most efficient method of computing
+    a determinant, it is a simple one, and it has the advantage of
+    being division free. To improve efficiency, this function uses
+    caching to avoid recomputing determinants of submatrices.
+    """
+    if not M.is_square:
+        raise NonSquareMatrixError()
+    if M.shape[0] == 0:
+        return M.one
+        # sympy/matrices/tests/test_matrices.py contains a test that
+        # suggests that the determinant of a 0 x 0 matrix is one, by
+        # convention.
+    return __det_laplace(M.as_immutable())
+
+
+def _det_bird(M):
+    r"""Compute the determinant of a matrix using Bird's algorithm.
+
+    Bird's algorithm is a simple division-free algorithm for computing, which
+    is of lower order than the Laplace's algorithm. It is described in [1]_.
+
+    References
+    ==========
+
+    .. [1] Bird, R. S. (2011). A simple division-free algorithm for computing
+           determinants. Inf. Process. Lett., 111(21), 1072-1074. doi:
+           10.1016/j.ipl.2011.08.006
+    """
+    def mu(X):
+        n = X.shape[0]
+        zero = X.domain.zero
+
+        total = zero
+        diag_sums = [zero]
+        for i in reversed(range(1, n)):
+            total -= X[i][i]
+            diag_sums.append(total)
+        diag_sums = diag_sums[::-1]
+
+        elems = [[zero] * i + [diag_sums[i]] + X_i[i + 1:] for i, X_i in
+                 enumerate(X)]
+        return DDM(elems, X.shape, X.domain)
+
+    Mddm = M._rep.to_ddm()
+    n = M.shape[0]
+    if n == 0:
+        return M.one
+        # sympy/matrices/tests/test_matrices.py contains a test that
+        # suggests that the determinant of a 0 x 0 matrix is one, by
+        # convention.
+    Fn1 = Mddm
+    for _ in range(n - 1):
+        Fn1 = mu(Fn1).matmul(Mddm)
+    detA = Fn1[0][0]
+    if n % 2 == 0:
+        detA = -detA
+
+    return Mddm.domain.to_sympy(detA)
+
+
 def _minor(M, i, j, method="berkowitz"):
     """Return the (i,j) minor of ``M``.  That is,
     return the determinant of the matrix obtained by deleting
@@ -836,7 +955,7 @@ def _minor(M, i, j, method="berkowitz"):
 
     method : string, optional
         Method to use to find the determinant of the submatrix, can be
-        "bareiss", "berkowitz" or "lu".
+        "bareiss", "berkowitz", "bird", "laplace" or "lu".
 
     Examples
     ========
